from numpy._core.records import (
    record as record,
    recarray as recarray,
    format_parser as format_parser,
    fromarrays as fromarrays,
    fromrecords as fromrecords,
    fromstring as fromstring,
    fromfile as fromfile,
    array as array
)

__all__: list[str]
__path__: list[str]
